/**
 *
 * SIROCCO
 * Copyright (C) 2011 France Telecom
 * Contact: sirocco@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *  $Id: ClusterManagerImpl.java 365 2011-09-30 13:31:13Z chou $
 *
 */

package org.ow2.sirocco.cloudmanager.clustermanager.impl;

import java.io.IOException;
import java.io.InputStream;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Date;
import java.util.Dictionary;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.NoSuchElementException;
import java.util.Set;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

import javax.jms.ObjectMessage;
import javax.jms.Session;
import javax.jms.Topic;
import javax.jms.TopicConnection;
import javax.jms.TopicConnectionFactory;
import javax.jms.TopicPublisher;
import javax.jms.TopicSession;
import javax.management.JMX;
import javax.management.MBeanServerConnection;
import javax.management.MalformedObjectNameException;
import javax.management.Notification;
import javax.management.NotificationEmitter;
import javax.management.NotificationListener;
import javax.management.ObjectName;
import javax.management.openmbean.CompositeData;
import javax.management.remote.JMXConnectionNotification;
import javax.management.remote.JMXConnector;
import javax.management.remote.JMXConnectorFactory;
import javax.management.remote.JMXServiceURL;
import javax.naming.InitialContext;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;

import org.osgi.service.cm.ConfigurationException;
import org.osgi.service.cm.ManagedService;
import org.ow2.sirocco.cloudmanager.clustermanager.api.AllocationMode;
import org.ow2.sirocco.cloudmanager.clustermanager.api.Domain;
import org.ow2.sirocco.cloudmanager.clustermanager.api.Host;
import org.ow2.sirocco.cloudmanager.clustermanager.api.IClusterManager;
import org.ow2.sirocco.cloudmanager.clustermanager.api.PerfCompositeMetric;
import org.ow2.sirocco.cloudmanager.clustermanager.api.PerfMetricBase;
import org.ow2.sirocco.cloudmanager.clustermanager.api.PerfMetricInfo;
import org.ow2.sirocco.cloudmanager.clustermanager.api.PerfSimpleMetric;
import org.ow2.sirocco.cloudmanager.clustermanager.api.ResourcePartitionInfo;
import org.ow2.sirocco.cloudmanager.clustermanager.api.ServerPool;
import org.ow2.sirocco.cloudmanager.clustermanager.api.VirtualMachineCreationSpec;
import org.ow2.sirocco.cloudmanager.clustermanager.api.VolumeSpec;
import org.ow2.sirocco.cloudmanager.clustermanager.api.event.NewPerfMetricsEvent;
import org.ow2.sirocco.cloudmanager.clustermanager.api.event.ResourceUsageEvent;
import org.ow2.sirocco.cloudmanager.clustermanager.api.event.VMImageStatusChangeEvent;
import org.ow2.sirocco.cloudmanager.clustermanager.api.event.VMImageUploadCompletionEvent;
import org.ow2.sirocco.cloudmanager.clustermanager.api.event.VirtualMachineCreatedEvent;
import org.ow2.sirocco.cloudmanager.clustermanager.api.event.VirtualMachineFailureEvent;
import org.ow2.sirocco.cloudmanager.clustermanager.api.event.VirtualMachineGuestOSReadyEvent;
import org.ow2.sirocco.cloudmanager.clustermanager.api.event.VirtualMachineMigrationEvent;
import org.ow2.sirocco.cloudmanager.clustermanager.api.event.VirtualMachinePowerStateChangeEvent;
import org.ow2.sirocco.cloudmanager.clustermanager.api.event.VolumeReadyEvent;
import org.ow2.sirocco.vmm.api.DiskOperation;
import org.ow2.sirocco.vmm.api.DomainMXBean;
import org.ow2.sirocco.vmm.api.GuestInfo;
import org.ow2.sirocco.vmm.api.HostMXBean;
import org.ow2.sirocco.vmm.api.NotificationTypes;
import org.ow2.sirocco.vmm.api.ResourcePartitionMXBean;
import org.ow2.sirocco.vmm.api.ResourceUsage;
import org.ow2.sirocco.vmm.api.ResourceUsage.DiskStats;
import org.ow2.sirocco.vmm.api.ResourceUsage.NetworkStats;
import org.ow2.sirocco.vmm.api.ServerPoolMXBean;
import org.ow2.sirocco.vmm.api.StoragePoolMXBean;
import org.ow2.sirocco.vmm.api.VMMAgentMXBean;
import org.ow2.sirocco.vmm.api.VMMException;
import org.ow2.sirocco.vmm.api.VNICSpec;
import org.ow2.sirocco.vmm.api.VNICSpec.MacAddressAssignement;
import org.ow2.sirocco.vmm.api.VirtualCdrom;
import org.ow2.sirocco.vmm.api.VirtualCdromSpec;
import org.ow2.sirocco.vmm.api.VirtualDisk;
import org.ow2.sirocco.vmm.api.VirtualDiskSpec;
import org.ow2.sirocco.vmm.api.VirtualMachineConfigSpec;
import org.ow2.sirocco.vmm.api.VirtualMachineImageMXBean;
import org.ow2.sirocco.vmm.api.VirtualMachineImageStoreMXBean;
import org.ow2.sirocco.vmm.api.VirtualMachineMXBean;
import org.ow2.sirocco.vmm.api.VirtualMachineMXBean.PowerState;
import org.ow2.sirocco.vmm.api.Volume;
import org.ow2.sirocco.vmm.api.monitoring.MonitorableMXBean;
import org.ow2.sirocco.vmm.api.monitoring.MonitorableMXBean.ConsolidationFunction;
import org.ow2.sirocco.vmm.api.monitoring.PerfMetric;
import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;

@SuppressWarnings("deprecation")
public class ClusterManagerImpl implements IClusterManager, ManagedService, NotificationListener {
    private static Log logger = LogFactory.getLog(ClusterManagerImpl.class);

    private static final String RESOURCE_PARTITION_USAGE_PROPERTY = "usage";

    private static final String RESOURCE_PARTITION_ON_DEMAND_USAGE = "OD";

    private static final String RESOURCE_PARTITION_ADVANCED_RESERVATION_USAGE = "AR";

    private static final String TOPIC_CONNECTION_FACTORY_NAME = "JTCF";

    private static final String TOPIC_NAME = "ClusterManager";

    private static final String PATH_CONFIG_MONITORING = "/META-INF/monitoring-config.xml";

    private static final String DEFAULT_AGENT_URL = "service:jmx:rmi:///jndi/rmi://localhost:9999/server";

    private static final int THREADPOOL_SIZE = 10;

    private static final int OBTAIN_GUEST_IP_ADDRESS_DEFAULT_WAITING_TIME_IN_SECONDS = 60 * 2;

    private static ExecutorService executorService = Executors.newFixedThreadPool(ClusterManagerImpl.THREADPOOL_SIZE);

    private MBeanServerConnection mbsc;

    private JMXConnector jmxc;

    private String connectionId;

    private boolean isConnected;

    private String jmxURL;

    private DomainMXBean rootDomainMBean;

    private Domain rootDomain;

    private Hashtable<ObjectName, HostMXBean> hostMBeanTable = new Hashtable<ObjectName, HostMXBean>();

    private Hashtable<String, VirtualMachineMXBean> vmMBeanTable = new Hashtable<String, VirtualMachineMXBean>();

    private Hashtable<String, ServerPoolMXBean> poolMBeanTable = new Hashtable<String, ServerPoolMXBean>();

    private Hashtable<String, StoragePoolMXBean> storagePoolMBeanTable = new Hashtable<String, StoragePoolMXBean>();

    private List<ResourcePartitionInfo> onDemandResourcePartitions = new ArrayList<ResourcePartitionInfo>();

    private List<ResourcePartitionInfo> advancedReservationResourcePartitions = new ArrayList<ResourcePartitionInfo>();

    private List<String> locations;

    // Monitoring attributs

    private PerfMetricInfo[] virtualMachinePerfMetricInfos = null;

    private PerfMetricInfo[] hostPerfMetricInfos = null;

    private PerfMetricInfo[] serverPoolPerfMetricInfos = null;

    private PerfMetricInfo[] domainPerfMetricInfos = null;

    private PerfMetricInfo[] userPerfMetricInfos = null;

    private PerfMetricInfo[] projectPerfMetricInfos = null;

    private long monitoring_period = 0;

    private long monitoring_delay = 0;

    private boolean monitoring_status = false;

    private boolean monitoring_init = false;

    private class VolumeInfo {
        StoragePoolMXBean storagePool;

        String volumeKey;

        Volume volume;

        VolumeInfo(final String providerId) throws VMMException {
            int i = providerId.indexOf(" ");
            if (i <= 0 || providerId.length() < i + 2) {
                throw new VMMException("Wrong providerId");
            }
            String storagePoolName = providerId.substring(0, i);
            this.storagePool = ClusterManagerImpl.this.getStoragePoolMXBean(storagePoolName);
            if (this.storagePool == null) {
                throw new VMMException("Cannot find storage pool " + storagePoolName);
            }
            this.volumeKey = providerId.substring(i + 1);
            this.volume = this.storagePool.getVolumeByKey(this.volumeKey);
            if (this.volume == null) {
                throw new VMMException("Cannot find volume " + this.volumeKey + " in storage pool " + storagePoolName);
            }
        }
    }

    public ClusterManagerImpl() {
    }

    @SuppressWarnings("rawtypes")
    @Override
    public void updated(final Dictionary properties) throws ConfigurationException {
        if (properties == null) {
            ClusterManagerImpl.logger.error("No config file !");
        } else {
            this.jmxURL = (String) properties.get("agent.url");
            if (this.jmxURL == null) {
                this.jmxURL = ClusterManagerImpl.DEFAULT_AGENT_URL;
            }
            ClusterManagerImpl.logger.info("Agent URL set to " + this.jmxURL);
        }
    }

    public synchronized void start() {
        ClusterManagerImpl.logger.debug("ClusterManagerImpl started");

    }

    public synchronized void shutdown() {
        ClusterManagerImpl.logger.debug("ClusterManagerImpl shutting down");
        if (this.isConnected) {
            this.disconnect();
        }
        ClusterManagerImpl.executorService.shutdownNow();
    }

    private void emitMessage(final Serializable payload) {
        try {
            InitialContext ctx = new InitialContext();
            TopicConnectionFactory topicConnectionFactory = (TopicConnectionFactory) ctx
                .lookup(ClusterManagerImpl.TOPIC_CONNECTION_FACTORY_NAME);
            TopicConnection connection = topicConnectionFactory.createTopicConnection();
            TopicSession session = connection.createTopicSession(false, Session.AUTO_ACKNOWLEDGE);
            Topic cloudAdminTopic = (Topic) ctx.lookup(ClusterManagerImpl.TOPIC_NAME);
            TopicPublisher topicPublisher = session.createPublisher(cloudAdminTopic);
            ObjectMessage message = session.createObjectMessage();
            message.setObject(payload);
            topicPublisher.publish(message);
            ClusterManagerImpl.logger.info("EMITTED EVENT " + payload.toString() + " on ClusterManager topic");
            topicPublisher.close();
            session.close();
            connection.close();
        } catch (Exception ex) {
            ClusterManagerImpl.logger.error("Failed to emit message", ex);
        }
    }

    private synchronized void checkConnection() {
        if (!this.isConnected) {
            this.connect();
        }
    }

    private void connect() {
        try {
            ClusterManagerImpl.logger.info("Connecting to " + this.jmxURL);
            JMXServiceURL url = new JMXServiceURL(this.jmxURL);
            this.jmxc = JMXConnectorFactory.connect(url, null);
            this.jmxc.addConnectionNotificationListener(this, null, null);
            this.connectionId = this.jmxc.getConnectionId();
            this.mbsc = this.jmxc.getMBeanServerConnection();
            Set<ObjectName> s = this.mbsc.queryNames(new ObjectName("org.ow2.sirocco.vmm.api:type=Agent"), null);
            ObjectName bridgeObjectName = s.toArray(new ObjectName[1])[0];
            VMMAgentMXBean bridgeMBean = JMX.newMXBeanProxy(this.mbsc, bridgeObjectName, VMMAgentMXBean.class);
            this.rootDomainMBean = bridgeMBean.getRootDomain();
            for (ResourcePartitionInfo rpInfo : this.onDemandResourcePartitions) {
                rpInfo.setMbean(JMX.newMXBeanProxy(this.mbsc, rpInfo.getObjectName(), ResourcePartitionMXBean.class));
            }
            for (ResourcePartitionInfo rpInfo : this.advancedReservationResourcePartitions) {
                rpInfo.setMbean(JMX.newMXBeanProxy(this.mbsc, rpInfo.getObjectName(), ResourcePartitionMXBean.class));
            }

            if (this.monitoring_period == 0) {
                this.initMonitoringConfig();
            }
            if (this.monitoring_status == true && this.monitoring_init == false) {
                this.monitoring_init = true;
                ClusterManagerImpl.logger.info("Init monitoring: ");

                ClusterManagerImpl.executorService.execute(new Runnable() {
                    public void run() {
                        Date startTime;
                        Date endTime;
                        long cpt = 0;
                        long pgcdInterval = 0;

                        ClusterManagerImpl.logger.info("Init monitoring : Thread launch");

                        for (PerfMetricInfo info : ClusterManagerImpl.this.virtualMachinePerfMetricInfos) {
                            if (cpt == 0) {
                                pgcdInterval = info.getInterval();
                                cpt++;
                            } else {
                                pgcdInterval = this.pgcd(pgcdInterval, info.getInterval());
                            }
                        }
                        cpt = 0;
                        ClusterManagerImpl.logger.info("PGCD:" + pgcdInterval);
                        while (ClusterManagerImpl.this.monitoring_status == true) {
                            HashMap<String, HashMap<PerfMetricInfo, PerfMetric[]>> tmp = new HashMap<String, HashMap<PerfMetricInfo, PerfMetric[]>>();
                            try {
                                long endTimeTmp;
                                String uuid = null;
                                Thread.sleep(1000 * ClusterManagerImpl.this.monitoring_period);
                                endTimeTmp = new Date().getTime();

                                for (VirtualMachineMXBean vm : ClusterManagerImpl.this.vmMBeanTable.values()) {
                                    // ClusterManagerImpl.logger.info("NOW-2mins:"
                                    // + new Date(new Date().getTime() -
                                    // 500000L)
                                    // + " : VM CREATION:" + vm.getStartTime());
                                    // if (vm != null && vm.getStartTime() !=
                                    // null
                                    // && new Date(new Date().getTime() -
                                    // 300000L).compareTo(vm.getStartTime()) >
                                    // 0) {
                                    synchronized (this) {
                                        if (vm != null) {
                                            uuid = vm.getUuid();
                                        }
                                    }
                                    tmp.put(uuid, new HashMap<PerfMetricInfo, PerfMetric[]>());

                                    HashMap<PerfMetricInfo, PerfMetric[]> tmp2 = tmp.get(uuid);
                                    for (PerfMetricInfo info : ClusterManagerImpl.this.virtualMachinePerfMetricInfos) {
                                        // now - monitoring delay
                                        endTime = new Date((endTimeTmp - (ClusterManagerImpl.this.monitoring_delay * 1000L))
                                            / (info.getInterval() * 1000L) * (info.getInterval() * 1000L));

                                        if (cpt > 0
                                            && (info.getInterval() <= ClusterManagerImpl.this.monitoring_period || (cpt
                                                % info.getInterval() == 0))) {

                                            if (info.getInterval() <= ClusterManagerImpl.this.monitoring_period) {
                                                startTime = new Date(endTime.getTime()
                                                    - (ClusterManagerImpl.this.monitoring_period * 1000L));
                                            } else {
                                                startTime = new Date(endTime.getTime() - (info.getInterval() * 1000L));
                                                // startTime = new Date(
                                                // endTime.getTime()
                                                // -
                                                // Math.abs((info.getInterval()
                                                // * 1000L -
                                                // ClusterManagerImpl.this.monitoring_period
                                                // * 1000L)));
                                            }

                                            synchronized (this) {
                                                if (vm != null) {
                                                    MonitorableMXBean.ConsolidationFunction cf = ConsolidationFunction
                                                        .fromValue(info.getType());
                                                    if (cf != null) {
                                                        if (cf.equals(MonitorableMXBean.ConsolidationFunction.AVERAGE)) {
                                                            ClusterManagerImpl.logger.info(info.getName() + "[" + startTime
                                                                + " : " + endTime + "][" + info.getInterval() + "], Interval="
                                                                + info.getInterval());
                                                        }
                                                        tmp2.put(
                                                            info,
                                                            vm.getPerfMetrics(info.getName(), startTime, endTime,
                                                                info.getInterval(), cf));
                                                    }

                                                }
                                            }
                                        }
                                    }
                                }
                                if (tmp != null && tmp.size() > 0) {
                                    ClusterManagerImpl.this.emitMessage(new NewPerfMetricsEvent(tmp));
                                }
                                cpt += ClusterManagerImpl.this.monitoring_period;
                            } catch (InterruptedException e) {
                                ClusterManagerImpl.logger.error("Interreupted : " + e);
                            } catch (VMMException e) {
                                ClusterManagerImpl.logger.error("VMM : " + e);
                            }
                        }
                        ClusterManagerImpl.logger.info("Close monitoring thread");
                        ClusterManagerImpl.this.monitoring_init = false;
                        ClusterManagerImpl.this.monitoring_period = 0;
                        ClusterManagerImpl.this.isConnected = false;

                    }

                    private long pgcd(final long a, final long b) {
                        return (b == 0) ? a : this.pgcd(b, a % b);
                    }

                });
            }
            this.isConnected = true;
        } catch (Exception ex) {
            ClusterManagerImpl.logger.error("ERROR : impossible to connect to JMX bridge: " + ex.getMessage());
        }
    }

    private void disconnect() {
        // XX should remove all notif listeners
        try {
            ClusterManagerImpl.logger.info("called");
            this.jmxc.removeConnectionNotificationListener(this);
            this.jmxc.close();
            this.monitoring_status = false;
            this.monitoring_init = false;
            this.monitoring_period = 0;
            this.hostMBeanTable.clear();
            this.vmMBeanTable.clear();
        } catch (Exception ex) {
            ClusterManagerImpl.logger.error("Disconnect error:", ex);
        }
    }

    public synchronized void handleNotification(final Notification notification, final Object handback) {
        String nType = notification.getType();

        try {
            if (nType.equals("jmx.remote.connection.closed")) {
                JMXConnectionNotification jcNotif = (JMXConnectionNotification) notification;
                if (jcNotif.getConnectionId().equals(this.connectionId)) {
                    ClusterManagerImpl.logger.warn("Connection lost with JMX Bridge " + this.jmxURL);
                    this.monitoring_status = false;
                    this.hostMBeanTable.clear();
                    this.vmMBeanTable.clear();
                    this.isConnected = false;
                }

            } else if (nType.equals(NotificationTypes.VM_MIGRATION)) {
                ObjectName vmName = (ObjectName) notification.getSource();
                String uuid = vmName.getKeyProperty("uuid");
                VirtualMachineMXBean vm = this.getVirtualMachineMBean(uuid);
                ClusterManagerImpl.logger.info("VM migration  name=" + vm.getNameLabel() + "to machine "
                    + notification.getMessage());
                String[] hosts = (String[]) notification.getUserData();
                this.emitMessage(new VirtualMachineMigrationEvent(uuid, hosts[0], hosts[1]));
                this.updateDataCenterResourceUsage();
            } else if (nType.equals(NotificationTypes.VM_STATE_CHANGE)) {
                ObjectName vmName = (ObjectName) notification.getSource();
                String vmPowerState = notification.getMessage();
                String uuid = vmName.getKeyProperty("uuid");
                VirtualMachineMXBean vm = this.getVirtualMachineMBean(uuid);
                if (vm != null) {
                    ClusterManagerImpl.logger.info("Notification: VM power state change: new state=" + vmPowerState + " vm="
                        + vm.getNameLabel());
                    this.emitMessage(new VirtualMachinePowerStateChangeEvent(vm.getNameLabel(), uuid, PowerState
                        .valueOf(vmPowerState)));
                }
            } else if (nType.equals(NotificationTypes.VM_ADD)) {
                ObjectName vmName = (ObjectName) notification.getUserData();
                String uuid = vmName.getKeyProperty("uuid");
                VirtualMachineMXBean vm = this.getVirtualMachineMBean(uuid);
                if (vm != null) {
                    ClusterManagerImpl.logger.info("Notification: NEW VM vm=" + vm.getNameLabel());
                }
                this.updateDataCenterResourceUsage();
            } else if (nType.equals(NotificationTypes.VM_DEL)) {
                ObjectName vmName = (ObjectName) notification.getUserData();
                String uuid = vmName.getKeyProperty("uuid");
                VirtualMachineMXBean vm = this.getVirtualMachineMBean(uuid);
                if (vm != null) {
                    ClusterManagerImpl.logger.info("Notification: DEL VM vm=" + vm.getNameLabel());
                }
                this.updateDataCenterResourceUsage();
            } else if (nType.equals(NotificationTypes.PERF_REPORT)) {
                // VirtualMachineManager.log.info("Received Perf report notif");
                ObjectName host = (ObjectName) notification.getSource();
                @SuppressWarnings("unchecked")
                Map<String, Object> map = (Map<String, Object>) notification.getUserData();
                this.updatePerfMetricsBuffers(host, map);
            }
        } catch (VMMException ex) {
            ClusterManagerImpl.logger.error("In JMX notif handler", ex);
        }
    }

    public VirtualMachineMXBean getVirtualMachineMBean(final String providerId) {
        this.checkConnection();
        VirtualMachineMXBean vmMBean = this.vmMBeanTable.get(providerId);
        if (vmMBean == null) {
            synchronized (this) {
                try {

                    Set<ObjectName> s = this.mbsc.queryNames(new ObjectName("org.ow2.sirocco.vmm.api:type=VirtualMachine,uuid="
                        + providerId + ",*"), null);
                    ObjectName vmMBeanName = s.toArray(new ObjectName[1])[0];

                    if (s.isEmpty()) {
                        return null;
                    }
                    vmMBean = JMX.newMXBeanProxy(this.mbsc, vmMBeanName, VirtualMachineMXBean.class, true);
                    ((NotificationEmitter) vmMBean).addNotificationListener(this, null, null);
                } catch (Exception ex) {
                    ClusterManagerImpl.logger.info("Error : ", ex);
                    return null;
                }
                this.vmMBeanTable.put(providerId, vmMBean);
            }
        }
        return vmMBean;
    }

    private StoragePoolMXBean getStoragePoolMXBean(final String poolName) {
        this.checkConnection();
        StoragePoolMXBean storagePool = this.storagePoolMBeanTable.get(poolName);
        if (storagePool == null) {
            synchronized (this) {
                try {

                    Set<ObjectName> s = this.mbsc.queryNames(new ObjectName("org.ow2.sirocco.vmm.api:type=StoragePool,name="
                        + poolName + ",*"), null);
                    ObjectName objectName = s.toArray(new ObjectName[1])[0];

                    if (s.isEmpty()) {
                        return null;
                    }
                    storagePool = JMX.newMXBeanProxy(this.mbsc, objectName, StoragePoolMXBean.class, true);
                } catch (Exception ex) {
                    ClusterManagerImpl.logger.info("Error : ", ex);
                    return null;
                }
                this.storagePoolMBeanTable.put(poolName, storagePool);
            }
        }
        return storagePool;
    }

    synchronized private void addHostMBean(final HostMXBean hostMBean, final ObjectName objectName) {
        try {
            NotificationEmitter ne = (NotificationEmitter) JMX.newMXBeanProxy(this.mbsc, objectName, Cloneable.class, true);
            ne.addNotificationListener(this, null, null);
            ClusterManagerImpl.logger.info("Added notification listener for host " + objectName);
        } catch (Exception ex) {
            ClusterManagerImpl.logger.error("Failed to add listener to HostMXBean", ex);
        }
        this.hostMBeanTable.put(objectName, hostMBean);
    }

    private HostMXBean getHostMBean(final ObjectName objectName) {
        this.checkConnection();
        HostMXBean hostMBean = this.hostMBeanTable.get(objectName);
        if (hostMBean == null) {
            hostMBean = JMX.newMXBeanProxy(this.mbsc, objectName, HostMXBean.class);
            this.addHostMBean(hostMBean, objectName);
        }
        return hostMBean;
    }

    private HostMXBean getHostMBean(final String hostName) {
        if (hostName == null) {
            return null;
        }
        this.checkConnection();
        try {
            // HostMXBean host = this.hostMBeanTable.get(new
            // ObjectName("org.ow2.sirocco.vmm.api:type=Host,name=" + hostName
            // + ",hostname=" + hostName.substring(hostName.lastIndexOf('/') +
            // 1, hostName.length())));
            // ClusterManagerImpl.logger.info("org.ow2.sirocco.vmm.api:type=Host,name="
            // + hostName + ",hostname="
            // + hostName.substring(hostName.lastIndexOf('/'),
            // hostName.length()));
            Set<ObjectName> s = this.mbsc.queryNames(
                new ObjectName("org.ow2.sirocco.vmm.api:type=Host,name=" + hostName + ",*"), null);
            ObjectName oname = s.toArray(new ObjectName[1])[0];
            HostMXBean host = JMX.newMXBeanProxy(this.mbsc, oname, HostMXBean.class, true);
            return host;
        } catch (MalformedObjectNameException e) {
            ClusterManagerImpl.logger.info(e);
            return null;
        } catch (NullPointerException e) {
            ClusterManagerImpl.logger.info(e);
            return null;
        } catch (IOException e) {
            ClusterManagerImpl.logger.info(e);
            return null;
        }
    }

    private ServerPoolMXBean getServerPoolMBean(final String objectNameString) {
        this.checkConnection();
        ServerPoolMXBean poolMBean = this.poolMBeanTable.get(objectNameString);
        ClusterManagerImpl.logger.debug("\n\n\n" + objectNameString);
        if (poolMBean == null) {
            try {
                poolMBean = JMX.newMXBeanProxy(this.mbsc, new ObjectName(objectNameString), ServerPoolMXBean.class);
                this.poolMBeanTable.put(objectNameString, poolMBean);
            } catch (Exception ex) {
                ClusterManagerImpl.logger.error("Failed to retrieve ServerPoolMXBean proxy for " + objectNameString, ex);
            }
        }
        return poolMBean;
    }

    private ServerPoolMXBean getServerPoolMBean2(final String serverPoolName) {
        if (serverPoolName == null) {
            return null;
        }
        this.checkConnection();
        try {

            // ServerPoolMXBean sp =
            // this.poolMBeanTable.get("org.ow2.sirocco.vmm.api:type=ServerPool,name="
            // + serverPoolName);
            //
            // ClusterManagerImpl.logger.info("org.ow2.sirocco.vmm.api:type=ServerPool,name="
            // + serverPoolName);
            // ClusterManagerImpl.logger.info(sp.getObjectName());
            // if (sp != null) {
            // sp = JMX.newMXBeanProxy(this.mbsc, sp.getObjectName(),
            // ServerPoolMXBean.class, true);
            // }
            Set<ObjectName> s = this.mbsc.queryNames(new ObjectName("org.ow2.sirocco.vmm.api:type=ServerPool,name="
                + serverPoolName + ",*"), null);
            ObjectName oname = s.toArray(new ObjectName[1])[0];
            ServerPoolMXBean sp = JMX.newMXBeanProxy(this.mbsc, oname, ServerPoolMXBean.class, true);
            return sp;
        } catch (NullPointerException e) {
            ClusterManagerImpl.logger.info(e);
            return null;
        } catch (MalformedObjectNameException e) {
            ClusterManagerImpl.logger.info(e);
            return null;
        } catch (IOException e) {
            ClusterManagerImpl.logger.info(e);
            return null;
        }
    }

    @Override
    public void createVirtualMachine(final VirtualMachineCreationSpec creationSpec, final Map<String, String> constraints,
        final AllocationMode allocationMode, final boolean startVM) throws VMMException {
        this.checkConnection();

        final ResourcePartitionMXBean resourcePartition;
        resourcePartition = this.getResourcePartition(allocationMode, constraints.get("hypervisor"),
            constraints.get("location"));
        if (resourcePartition == null) {
            String errorMessage = "Cannot find resource partition for " + allocationMode + " , hypervisor="
                + constraints.get("hypervisor") + " location=" + constraints.get("location");
            ClusterManagerImpl.logger.error(errorMessage);
            throw new VMMException(errorMessage);
        }

        final VirtualMachineConfigSpec configSpec = new VirtualMachineConfigSpec();
        configSpec.setName(creationSpec.getName());
        configSpec.setNumVCPUs(creationSpec.getNumVCPUs());
        configSpec.setMemoryMB(creationSpec.getMemorySizeMB());

        VirtualDiskSpec diskSpec = new VirtualDiskSpec();
        diskSpec.setCapacityMB(creationSpec.getDiskCapacityMB());
        diskSpec.setDiskOp(DiskOperation.CREATE_FROM);
        // imageId is built as: <storage pool name> <volume key>
        int i = creationSpec.getVmImageId().indexOf(" ");
        if (i <= 0 || creationSpec.getVmImageId().length() < i + 2) {
            throw new VMMException("Wrong providerId");
        }
        String storagePoolName = creationSpec.getVmImageId().substring(0, i);
        String volumeKey = creationSpec.getVmImageId().substring(i + 1);

        StoragePoolMXBean vmImageStoragePool = this.getStoragePoolMXBean(storagePoolName);
        if (vmImageStoragePool == null) {
            throw new VMMException("Cannot find storage pool " + storagePoolName + " referenced by VM Image");
        }
        Volume vmImageVolume = vmImageStoragePool.getVolumeByKey(volumeKey);
        if (vmImageVolume == null) {
            throw new VMMException("Cannot find volume with key " + volumeKey + " referenced by VM Image");
        }
        diskSpec.setVolume(vmImageVolume);

        // find the default storage pool of the server pool which will host the
        // VM
        ServerPoolMXBean serverPool = resourcePartition.getRootServerPool();
        StoragePoolMXBean targetStoragePool = null;
        for (StoragePoolMXBean storagePool : serverPool.getStoragePools()) {
            Map<String, String> props = storagePool.getProperties();
            if (props != null) {
                @SuppressWarnings("unused")
                String isDefault = props.get("default");
                targetStoragePool = storagePool;
                break;
            }
        }
        if (targetStoragePool == null) {
            targetStoragePool = serverPool.getStoragePools().get(0);
            ClusterManagerImpl.logger.warn("No default storage pool for server pool " + serverPool.getName()
                + ", defaulting to storage pool " + targetStoragePool.getName());
        }
        diskSpec.setStoragePool(targetStoragePool);
        configSpec.setDiskSpecs(Collections.singletonList(diskSpec));

        // add default empty cdrom drive
        VirtualCdromSpec cdromSpec = new VirtualCdromSpec();
        cdromSpec.setIsoVolume(null);
        configSpec.setCdromSpecs(Collections.singletonList(cdromSpec));

        VNICSpec vnicSpec = new VNICSpec();
        vnicSpec.setAddressType(MacAddressAssignement.GENERATED);
        vnicSpec.setNetworkName("default");
        configSpec.setVnicSpecs(Collections.singletonList(vnicSpec));

        Map<String, String> vmProps = new HashMap<String, String>();
        vmProps.put("bootDevice", "disk");
        configSpec.setProperties(vmProps);

        ClusterManagerImpl.executorService.execute(new Runnable() {
            @Override
            public void run() {
                ClusterManagerImpl.logger.info("Creating VM " + creationSpec.getName());
                VirtualMachineMXBean vmMBean = null;
                try {
                    vmMBean = resourcePartition.createVirtualMachine(configSpec, creationSpec.getCustomizationSpec(), startVM,
                        true);
                    if (startVM) {
                        ClusterManagerImpl.logger.info("Starting VM " + creationSpec.getName());
                    } else {
                        // nothing to do.
                    }
                } catch (VMMException ex) {
                    ClusterManagerImpl.this.emitMessage(new VirtualMachineFailureEvent(creationSpec.getName(), null, ex
                        .getClass().getName(), ex.getMessage()));
                    return;
                }

                try {
                    NotificationEmitter ne = (NotificationEmitter) JMX.newMXBeanProxy(ClusterManagerImpl.this.mbsc,
                        vmMBean.getObjectName(), Cloneable.class, true);
                    ne.addNotificationListener(ClusterManagerImpl.this, null, null);
                } catch (Exception ex) {
                    ClusterManagerImpl.logger.error("", ex);
                }
                try {
                    String uuid = vmMBean.getUuid();
                    ClusterManagerImpl.this.vmMBeanTable.put(uuid, vmMBean);
                    String hostName = vmMBean.getHostMBean().getHostName();
                    if (startVM) {
                        ClusterManagerImpl.logger.info("VM " + creationSpec.getName() + " created and started on host "
                            + hostName);
                        // retrieve guest OS IP address
                        int wait_time_sec = 0;
                        GuestInfo guestInfo = null;
                        while (wait_time_sec < ClusterManagerImpl.OBTAIN_GUEST_IP_ADDRESS_DEFAULT_WAITING_TIME_IN_SECONDS) {
                            guestInfo = vmMBean.getGuestInfo();
                            if (guestInfo != null && guestInfo.getIpAddresses() != null
                                && guestInfo.getIpAddresses().size() > 0) {
                                break;
                            }
                            try {
                                Thread.sleep(1000 * 15);
                            } catch (InterruptedException e) {
                            }
                            wait_time_sec += 15;
                        }
                        if (guestInfo == null || guestInfo.getIpAddresses() == null || guestInfo.getIpAddresses().size() == 0) {
                            ClusterManagerImpl.logger.warn("Failed to retrieve IP address of VM " + creationSpec.getName()
                                + " after " + ClusterManagerImpl.OBTAIN_GUEST_IP_ADDRESS_DEFAULT_WAITING_TIME_IN_SECONDS
                                + " seconds");
                        }
                        ClusterManagerImpl.this.emitMessage(new VirtualMachineGuestOSReadyEvent(creationSpec.getName(), uuid,
                            ClusterManagerImpl.this.virtualMachinePerfMetricInfos));
                    } else {
                        ClusterManagerImpl.logger.info("VM " + creationSpec.getName() + " created on host " + hostName);
                        ClusterManagerImpl.this.emitMessage(new VirtualMachineCreatedEvent(creationSpec.getName(), uuid));
                    }
                } catch (VMMException ex) {
                    ClusterManagerImpl.logger.error("While creating VM " + creationSpec.getName(), ex);
                }
            }
        });

    }

    @Override
    public synchronized void destroyVirtualMachine(final String providerId) throws VMMException {
        this.checkConnection();
        VirtualMachineMXBean vmMBean = this.getVirtualMachineMBean(providerId);
        if (vmMBean != null) {
            vmMBean.destroy();

            this.vmMBeanTable.remove(providerId);

        } else {
            ClusterManagerImpl.logger.error("cannot destroy VM with unknow providerId=" + providerId);
        }
    }

    @Override
    public void createImageFromVirtualMachine(final String providerId, final Integer vmiId, final String name,
        final String description) throws VMMException {
        this.checkConnection();
        final VirtualMachineMXBean vmMBean = this.getVirtualMachineMBean(providerId);
        if (vmMBean == null) {
            ClusterManagerImpl.logger.info("cannot find VM with given providerId");
            throw new VMMException("Cannot find VM with providerId=" + providerId);
        }
        ClusterManagerImpl.executorService.execute(new Runnable() {
            @Override
            public void run() {
                VMImageStatusChangeEvent event = new VMImageStatusChangeEvent();
                event.setImageId(vmiId);
                try {
                    StoragePoolMXBean storagePool = vmMBean.getHostMBean().getServerPool().getStoragePools().get(0);
                    Volume volume = vmMBean.makeImage(name, storagePool);
                    event.setImageProviderId(storagePool.getPath() + " " + volume.getKey());
                    event.setVmProviderId(providerId);
                    event.setImageStatus("AVAILABLE");
                } catch (VMMException ex) {
                    event.setImageStatus("ERROR");
                    event.setErrorDetail(ex.getClass().getName());
                }
                ClusterManagerImpl.this.emitMessage(event);
            }
        });
    }

    public void destroyVMImage(final String providerId) throws VMMException {
        this.checkConnection();
        this.destroyVMImage(providerId, this.rootDomainMBean);
    }

    private StoragePoolMXBean findStoragePoolMatchingFormat(final String format) throws VMMException {
        for (ResourcePartitionInfo rp : this.onDemandResourcePartitions) {
            for (StoragePoolMXBean pool : rp.getMbean().getRootServerPool().getStoragePools()) {
                for (String poolSupportFormat : pool.getSupportedDiskImageFormats()) {
                    if (poolSupportFormat.equalsIgnoreCase(format)) {
                        return pool;
                    }
                }
            }
        }
        return null;
    }

    private StoragePoolMXBean findStoragePool(final String hypervisor, final String location) throws VMMException {
        for (ResourcePartitionInfo rp : this.onDemandResourcePartitions) {
            if (location == null || (rp.getLocation() != null && rp.getLocation().equals(location))) {
                ServerPoolMXBean serverPool = rp.getMbean().getRootServerPool();
                if (hypervisor == null || serverPool.getHypervisor().equals(hypervisor)) {
                    for (StoragePoolMXBean pool : serverPool.getStoragePools()) {
                        String prop = pool.getProperties().get("default");
                        if (prop != null && prop.equalsIgnoreCase("true")) {
                            return pool;
                        }
                    }
                }
            }
        }
        return null;
    }

    @Override
    public void uploadVMImage(final String url, final String format, final String name, final String description,
        final String checkSum) throws VMMException {
        this.checkConnection();
        this.getResourceTree();
        final StoragePoolMXBean storagePool = this.findStoragePoolMatchingFormat(format);
        if (storagePool == null) {
            throw new VMMException("Cannot find storage pool supporting " + format + " format");
        }

        ClusterManagerImpl.executorService.execute(new Runnable() {
            @Override
            public void run() {
                try {
                    Volume volume = storagePool.uploadVolume(url, format, name, description, checkSum);
                    String providerId = storagePool.getPath() + " " + volume.getKey();
                    ClusterManagerImpl.this.emitMessage(new VMImageUploadCompletionEvent(url, providerId));
                } catch (VMMException ex) {
                    ClusterManagerImpl.logger.error("Failed to upload VM Image from URL=" + url, ex);
                }
            }
        });
    }

    private boolean destroyVMImage(final String providerId, final DomainMXBean domain) throws VMMException {

        for (DomainMXBean subDomain : domain.getSubDomains()) {
            if (this.destroyVMImage(providerId, subDomain)) {
                break;
            }
        }
        for (ServerPoolMXBean sp : domain.getServerPools()) {
            VirtualMachineImageStoreMXBean dStore = sp.getVMImageStore();
            VirtualMachineImageMXBean image;

            try {
                image = dStore.lookUpByUUID(providerId);
            } catch (Exception ex) {
                continue;
            }

            if (image != null) {
                dStore.removeVMImageTemplate(image);
                return true;
            }
        }
        return false;
    }

    @Override
    public String createVolume(final VolumeSpec spec, final Object handback) throws VMMException {
        // This is the volume's providerId.
        String result = null;

        this.checkConnection();
        this.getResourceTree();

        String hypervisor = spec.getConstraints() != null ? spec.getConstraints().get("hypervisor") : null;
        String location = spec.getConstraints() != null ? spec.getConstraints().get("location") : null;

        StoragePoolMXBean storagePool = this.findStoragePool(hypervisor, location);
        if (storagePool == null) {
            throw new VMMException("Cannot find storage pool matching hypervisor=" + hypervisor + " location=" + location);
        }
        String format = null;
        for (String f : storagePool.getSupportedDiskImageFormats()) {
            if (!f.equalsIgnoreCase("iso")) {
                format = f;
                break;
            }
        }
        Volume vol = storagePool.createVolume(spec.getName(), spec.getCapacityInMB(), format);
        String providerId = storagePool.getPath() + " " + vol.getKey();
        this.emitMessage(new VolumeReadyEvent(spec.getName(), providerId, handback));
        result = providerId;
        return result;
    }

    @Override
    public void destroyVolume(final String volProviderId) throws VMMException {
        this.checkConnection();
        VolumeInfo volInfo = new VolumeInfo(volProviderId);
        volInfo.storagePool.destroyVolume(volInfo.volumeKey);
    }

    @Override
    public void attachVolumeToVM(final String vmProviderId, final String volumeId) throws VMMException {
        this.checkConnection();
        VolumeInfo volInfo = new VolumeInfo(volumeId);
        VirtualMachineMXBean vm = this.getVirtualMachineMBean(vmProviderId);
        if (vm == null) {
            throw new VMMException("Cannot find VM " + vmProviderId);
        }
        if (volInfo.volume.getFormat().equals("iso")) {
            List<VirtualCdrom> cdroms = vm.getVirtualCdroms();
            if (cdroms == null || cdroms.size() == 0) {
                throw new VMMException("Cannot find CD-ROM drive in VM " + vm.getNameLabel());
            }
            vm.changeCdromMedia(cdroms.get(0).getDevice(), volInfo.volume);

        } else {
            VirtualDiskSpec diskSpec = new VirtualDiskSpec();
            diskSpec.setDiskOp(DiskOperation.ATTACH);
            diskSpec.setVolume(volInfo.volume);
            vm.attachVolume(diskSpec);
        }
    }

    @Override
    public void detachVolumeFromVM(final String vmProviderId, final String volumeId) throws VMMException {
        this.checkConnection();
        VolumeInfo volInfo = new VolumeInfo(volumeId);
        VirtualMachineMXBean vm = this.getVirtualMachineMBean(vmProviderId);
        if (vm == null) {
            throw new VMMException("Cannot find VM " + vmProviderId);
        }
        if (volInfo.volume.getFormat().equals("iso")) {
            List<VirtualCdrom> cdroms = vm.getVirtualCdroms();
            if (cdroms == null || cdroms.size() == 0) {
                throw new VMMException("Cannot find CD-ROM drive in VM " + vm.getNameLabel());
            }
            vm.changeCdromMedia(cdroms.get(0).getDevice(), null);
        } else {
            vm.detachVolume(volInfo.volumeKey);
        }
    }

    @Override
    public String[] getVMVolumes(final String vmProviderId) throws VMMException {
        VirtualMachineMXBean vm = this.getVirtualMachineMBean(vmProviderId);
        if (vm == null) {
            throw new VMMException("Cannot find VM " + vmProviderId);
        }
        List<VirtualDisk> disks = vm.getVirtualDisks();
        String[] result = new String[disks.size()];
        for (int i = 0; i < disks.size(); i++) {
            Volume vol = disks.get(i).getVolume();
            result[i] = vol.getStoragePool().getPath() + " " + vol.getKey();
        }
        return result;
    }

    @Override
    public PerfMetricBase[] getHostPerfMetrics(final String hostId, final String metricId, final Date startTime,
        final Date endTime) throws VMMException {
        if (!metricId.equals("cpu.load")) {
            throw new VMMException("InvalidPerfMetricIdException");
        }
        ObjectName objectName;
        try {
            objectName = new ObjectName(hostId);
        } catch (MalformedObjectNameException ex) {
            throw new VMMException("InvalidHostIdException");
        }
        CircularBuffer<HostUsage> hostBuffer = this.hostPerfMetricsMap.get(objectName);
        if (hostBuffer == null) {
            ClusterManagerImpl.logger.info("Cannot find perf data for Host " + hostId);
            return new PerfCompositeMetric[0];
        }
        List<PerfCompositeMetric> metrics = new ArrayList<PerfCompositeMetric>();
        synchronized (hostBuffer) {
            for (HostUsage hostUsage : hostBuffer) {
                // VirtualMachineManager.log.info("Sample time=" +
                // hostUsage.timeStamp);
                if (startTime == null || hostUsage.timeStamp.after(startTime)) {
                    if (endTime == null || hostUsage.timeStamp.before(endTime)) {
                        float cpuLoads[] = new float[hostUsage.keySet().size()];
                        String components[] = new String[hostUsage.keySet().size()];
                        int i = 0;
                        for (String vmLabel : hostUsage.keySet()) {
                            cpuLoads[i] = hostUsage.get(vmLabel).getCpuLoad();
                            components[i++] = vmLabel;
                        }
                        PerfCompositeMetric metric = new PerfCompositeMetric(hostUsage.timeStamp, cpuLoads, components);
                        metrics.add(metric);
                    }
                }
            }
        }
        return metrics.toArray(new PerfCompositeMetric[0]);
    }

    @Override
    public List<Host> getHostRefFromServerPool(final String arg0) {
        // TODO Auto-generated method stub
        return null;
    }

    public ResourcePartitionMXBean getResourcePartition(final AllocationMode allocationMode, final String hypervisorType,
        final String location) {
        this.checkConnection();
        this.getResourceTree();
        List<ResourcePartitionInfo> rpInfos = allocationMode == AllocationMode.ON_DEMAND ? this.onDemandResourcePartitions
            : this.advancedReservationResourcePartitions;
        for (ResourcePartitionInfo info : rpInfos) {
            if (!info.getHypervisor().equalsIgnoreCase(hypervisorType)) {
                continue;
            }
            if (location == null || location.equals(info.getLocation())) {
                return info.getMbean();
            }
        }
        return null;
    }

    @Override
    public List<ResourcePartitionInfo> getResourcePartitions() {
        this.checkConnection();
        this.getResourceTree();
        List<ResourcePartitionInfo> result = new ArrayList<ResourcePartitionInfo>(this.onDemandResourcePartitions);
        result.addAll(this.advancedReservationResourcePartitions);
        return result;
    }

    @Override
    public List<String> getLocations() {
        this.checkConnection();
        if (this.locations == null) {
            this.locations = new ArrayList<String>();
            this.locations.add("anywhere");
            this.collectLocations(this.rootDomainMBean, this.locations);
        }
        return this.locations;
    }

    private void collectLocations(final DomainMXBean domain, final List<String> locs) {
        String loc = domain.getAttribute("location");
        if (loc != null && !loc.equals("")) {
            locs.add(loc);
        }
        for (DomainMXBean subDomain : domain.getSubDomains()) {
            this.collectLocations(subDomain, locs);
        }
    }

    private Host buildHost(final HostMXBean hostMBean) {
        Host hostVO = new Host();
        hostVO.setId(hostMBean.getObjectName().toString());

        try {
            hostVO.setHostName(hostMBean.getHostName());
            hostVO.setNumCPUs(hostMBean.getNumCPU());
            hostVO.setCpuInfo(hostMBean.getCPUInfo());
            hostVO.setHypervisorInfo(hostMBean.getHypervisorInfo());
            int cpuSpeedMHz = Integer.parseInt(hostVO.getCpuInfo().get("speedMHz"));
            hostVO.setCpuCapacityMHz(cpuSpeedMHz * hostVO.getNumCPUs());
            hostVO.setCpuCoreCapacity(hostVO.getNumCPUs());
            hostVO.setMemoryCapacityMB((int) hostMBean.getTotalMemoryMB());
            hostVO.setStorageCapacityMB((int) hostMBean.getServerPool().getStorageCapacityMB());
        } catch (VMMException ex) {
            ClusterManagerImpl.logger.error("Cannot init HostVO", ex);
        }

        return hostVO;
    }

    private void walkResourcePartition(final ResourcePartitionMXBean partition, final String hypervisor, final String location) {
        String path = partition.getPath();
        if (ClusterManagerImpl.RESOURCE_PARTITION_ON_DEMAND_USAGE.equals(partition.getProperties().get(
            ClusterManagerImpl.RESOURCE_PARTITION_USAGE_PROPERTY))) {
            ResourcePartitionInfo rpInfo = new ResourcePartitionInfo(AllocationMode.ON_DEMAND, partition, path, hypervisor,
                location);
            this.onDemandResourcePartitions.add(rpInfo);
            ClusterManagerImpl.logger.info("Adding on-demand " + rpInfo);
        }
        if (ClusterManagerImpl.RESOURCE_PARTITION_ADVANCED_RESERVATION_USAGE.equals(partition.getProperties().get(
            ClusterManagerImpl.RESOURCE_PARTITION_USAGE_PROPERTY))) {
            ResourcePartitionInfo rpInfo = new ResourcePartitionInfo(AllocationMode.ADVANCED_RESERVATION, partition, path,
                hypervisor, location);
            this.advancedReservationResourcePartitions.add(rpInfo);
            ClusterManagerImpl.logger.info("Adding advanced reservation " + rpInfo);
        }
        for (ResourcePartitionMXBean child : partition.getSubResourcePartitions()) {
            this.walkResourcePartition(child, hypervisor, location);
        }
    }

    private ServerPool buildServerPool(final ServerPoolMXBean poolMBean, final String location) {
        ServerPool poolVO = new ServerPool();
        poolVO.setId(poolMBean.getObjectName().toString());
        poolVO.setName(poolMBean.getName());
        for (HostMXBean host : poolMBean.getManagedHosts()) {
            Host hostVO = this.buildHost(host);
            poolVO.addHost(hostVO);
            this.addHostMBean(host, host.getObjectName());
            poolVO.setCpuCapacityMHz(poolVO.getCpuCapacityMHz() + hostVO.getCpuCapacityMHz());
            poolVO.setCpuCoreCapacity(poolVO.getCpuCoreCapacity() + hostVO.getCpuCoreCapacity());
            poolVO.setMemoryCapacityMB(poolVO.getMemoryCapacityMB() + hostVO.getMemoryCapacityMB());
        }
        String hypervisor = poolMBean.getHypervisor();
        this.walkResourcePartition(poolMBean, hypervisor, location);
        poolVO.setStorageCapacityMB((int) poolMBean.getStorageCapacityMB());
        return poolVO;
    }

    private Domain buildDomain(final DomainMXBean domainMBean) {
        Domain domain = new Domain();
        domain.setId(domainMBean.getObjectName().toString());
        domain.setName(domainMBean.getName());
        String location = domainMBean.getAttribute("location");
        for (DomainMXBean subDomain : domainMBean.getSubDomains()) {
            Domain subDomainVO = this.buildDomain(subDomain);
            domain.addSubDomain(subDomainVO);
            domain.setCpuCapacityMHz(domain.getCpuCapacityMHz() + subDomainVO.getCpuCapacityMHz());
            domain.setCpuCoreCapacity(domain.getCpuCoreCapacity() + subDomainVO.getCpuCoreCapacity());
            domain.setMemoryCapacityMB(domain.getMemoryCapacityMB() + subDomainVO.getMemoryCapacityMB());
            domain.setStorageCapacityMB(domain.getStorageCapacityMB() + subDomainVO.getStorageCapacityMB());
        }
        for (ServerPoolMXBean pool : domainMBean.getServerPools()) {
            ServerPool poolVO = this.buildServerPool(pool, location);
            domain.addServerPool(poolVO);
            domain.setCpuCapacityMHz(domain.getCpuCapacityMHz() + poolVO.getCpuCapacityMHz());
            domain.setCpuCoreCapacity(domain.getCpuCoreCapacity() + poolVO.getCpuCoreCapacity());
            domain.setMemoryCapacityMB(domain.getMemoryCapacityMB() + poolVO.getMemoryCapacityMB());
            domain.setStorageCapacityMB(domain.getStorageCapacityMB() + poolVO.getStorageCapacityMB());
        }
        return domain;
    }

    private void updateHostResourceUsage(final Host host) {
        try {
            ObjectName objectName = new ObjectName(host.getId());
            HostMXBean hostMBean = this.getHostMBean(objectName);
            if (hostMBean != null) {
                int numCpuCoreAllocated = 0;
                for (VirtualMachineMXBean vm : hostMBean.getResidentVMs()) {
                    numCpuCoreAllocated += vm.getNumVCPUs();
                }
                host.setCpuCoreAllocated(numCpuCoreAllocated);
                host.setCpuAllocatedMHz(numCpuCoreAllocated * Integer.parseInt(host.getCpuInfo().get("speedMHz")));
                host.setMemoryAllocatedMB((int) (hostMBean.getTotalMemoryMB() - hostMBean.getFreeMemoryMB()));
                host.setStorageAllocatedMB((int) (hostMBean.getServerPool().getStorageCapacityMB() - hostMBean.getServerPool()
                    .getFreeStorageMBCapacity()));
            }
        } catch (Exception ex) {
            ClusterManagerImpl.logger.error("Cannot update HostVO resource usage", ex);
        }
    }

    private void updateServerPoolResourceUsage(final ServerPool pool) {
        int usedCpuCore = 0;
        int usedCpuMHz = 0;
        int usedStorageMB = 0;
        int usedMemoryMB = 0;
        for (Host host : pool.getHosts()) {
            this.updateHostResourceUsage(host);
            usedCpuCore += host.getCpuCoreAllocated();
            usedCpuMHz += host.getCpuAllocatedMHz();
            usedMemoryMB += host.getMemoryAllocatedMB();
        }
        ServerPoolMXBean poolMBean = this.getServerPoolMBean(pool.getId());
        if (poolMBean != null) {
            usedStorageMB = (int) (poolMBean.getStorageCapacityMB() - poolMBean.getFreeStorageMBCapacity());
        }
        pool.setCpuAllocatedMHz(usedCpuMHz);
        pool.setCpuCoreAllocated(usedCpuCore);
        pool.setMemoryAllocatedMB(usedMemoryMB);
        pool.setStorageAllocatedMB(usedStorageMB);
    }

    private void updateDomainResourceUsage(final Domain domain) {
        int usedCpuCore = 0;
        int usedCpuMHz = 0;
        int usedStorageMB = 0;
        int usedMemoryMB = 0;
        for (Domain subDomain : domain.getSubDomains()) {
            this.updateDomainResourceUsage(subDomain);
            usedCpuCore += subDomain.getCpuCoreAllocated();
            usedCpuMHz += subDomain.getCpuAllocatedMHz();
            usedMemoryMB += subDomain.getMemoryAllocatedMB();
            usedStorageMB += subDomain.getStorageAllocatedMB();
        }
        for (ServerPool pool : domain.getServerPools()) {
            this.updateServerPoolResourceUsage(pool);
            usedCpuCore += pool.getCpuCoreAllocated();
            usedCpuMHz += pool.getCpuAllocatedMHz();
            usedMemoryMB += pool.getMemoryAllocatedMB();
            usedStorageMB += pool.getStorageAllocatedMB();
        }
        domain.setCpuAllocatedMHz(usedCpuMHz);
        domain.setCpuCoreAllocated(usedCpuCore);
        domain.setMemoryAllocatedMB(usedMemoryMB);
        domain.setStorageAllocatedMB(usedStorageMB);
    }

    private synchronized void updateDataCenterResourceUsage() {
        this.checkConnection();
        if (this.rootDomain != null) {
            this.updateDomainResourceUsage(this.rootDomain);
            this.emitMessage(new ResourceUsageEvent(this.rootDomain));
        }
    }

    @Override
    public synchronized Domain getResourceTree() {
        if (this.rootDomain == null) {
            this.checkConnection();
            this.rootDomain = this.buildDomain(this.rootDomainMBean);
            this.updateDomainResourceUsage(this.rootDomain);
            ClusterManagerImpl.logger.info("DataCenter tree: \n" + this.rootDomain);
        }
        return this.rootDomain;
    }

    @Override
    public List<ServerPool> getServerPools() {
        // TODO Auto-generated method stub
        return null;
    }

    @Override
    public float getVirtualMachineCPULoad(final String providerId) {
        this.checkConnection();
        VirtualMachineMXBean vmMBean = this.getVirtualMachineMBean(providerId);
        if (vmMBean != null) {
            try {
                return vmMBean.getCPULoad();
            } catch (VMMException ex) {
                ClusterManagerImpl.logger.error("Cannot retrieve CPU load", ex);
            }
        }
        return 0;
    }

    @Override
    public String getVirtualMachineConsole(final String providerId) throws VMMException {
        this.checkConnection();
        VirtualMachineMXBean vmMBean = this.getVirtualMachineMBean(providerId);
        if (vmMBean == null) {
            ClusterManagerImpl.logger.info("getVirtualMachineConsole: cannot find VM with given providerId");
            return null;
        }
        return vmMBean.getConsole();
    }

    @Override
    public long getVirtualMachineMemoryUsedMB(final String providerId) {
        this.checkConnection();
        VirtualMachineMXBean vmMBean = this.getVirtualMachineMBean(providerId);
        if (vmMBean != null) {
            try {
                return vmMBean.getMemoryUsedMB();
            } catch (VMMException ex) {
                ClusterManagerImpl.logger.error("getMemoryUsedMB failed", ex);
            }
        }
        return 0;
    }

    private static int DEFAULT_RESOURCE_USAGE_CAPACITY = 60 * 5;

    private static class CircularBuffer<Item> implements Iterable<Item> {
        Item buf[];

        int head = 0;

        int tail = 0;

        @SuppressWarnings("unchecked")
        CircularBuffer(final int capacity) {
            this.buf = (Item[]) new Object[capacity];
        }

        synchronized void addItem(final Item item) {
            this.buf[this.tail++] = item;
            this.tail %= this.buf.length;
            if (this.tail == this.head) {
                this.head = (this.head + 1) % this.buf.length;
            }
        }

        public Iterator<Item> iterator() {
            return new CircularBufferIterator();
        }

        private class CircularBufferIterator implements Iterator<Item> {
            private int i = CircularBuffer.this.head;

            public boolean hasNext() {
                return this.i != CircularBuffer.this.tail;
            }

            public void remove() {
                throw new UnsupportedOperationException();
            }

            public Item next() {
                if (!this.hasNext()) {
                    throw new NoSuchElementException();
                }
                Item item = CircularBuffer.this.buf[this.i];
                this.i = (this.i + 1) % CircularBuffer.this.buf.length;
                return item;
            }
        }

    }

    private static class HostUsage extends HashMap<String, ResourceUsage> {

        private static final long serialVersionUID = -6026211251697850297L;

        Date timeStamp;

        @Override
        public String toString() {
            StringBuilder sb = new StringBuilder();
            sb.append("(time=" + this.timeStamp);
            for (String vmLabel : this.keySet()) {
                sb.append(",(vm=" + vmLabel + ",cpu=" + this.get(vmLabel).getCpuLoad() + ")");
            }
            return sb.toString();
        }
    }

    private Map<String, CircularBuffer<ResourceUsage>> virtualMachinePerfMetricsMap = new HashMap<String, CircularBuffer<ResourceUsage>>();

    private Map<ObjectName, CircularBuffer<HostUsage>> hostPerfMetricsMap = new HashMap<ObjectName, CircularBuffer<HostUsage>>();

    private void updatePerfMetricsBuffers(final ObjectName host, final Map<String, Object> map) {
        // VirtualMachineManager.log.info("updating PerfMetricsBuffer for host "
        // + host);
        HostUsage hostUsage = new HostUsage();
        for (String vmLabel : map.keySet()) {
            ResourceUsage vmUsage = ResourceUsage.from((CompositeData) map.get(vmLabel));
            // XXX
            int i = vmLabel.lastIndexOf("-");
            if (i != -1) {
                vmLabel = vmLabel.substring(0, i);
                hostUsage.put(vmLabel, vmUsage);
                hostUsage.timeStamp = vmUsage.getSamplingTime();
                CircularBuffer<ResourceUsage> vmBuffer = this.virtualMachinePerfMetricsMap.get(vmLabel);
                if (vmBuffer == null) {
                    vmBuffer = new CircularBuffer<ResourceUsage>(ClusterManagerImpl.DEFAULT_RESOURCE_USAGE_CAPACITY);
                    this.virtualMachinePerfMetricsMap.put(vmLabel, vmBuffer);
                    // ClusterManagerImpl.logger.info("Added Perf CircularBuffer for VM "
                    // + vmLabel);
                }
                vmBuffer.addItem(vmUsage);
            }
            if (map.keySet().size() > 0) {
                CircularBuffer<HostUsage> hostBuffer = this.hostPerfMetricsMap.get(host);
                if (hostBuffer == null) {
                    hostBuffer = new CircularBuffer<HostUsage>(ClusterManagerImpl.DEFAULT_RESOURCE_USAGE_CAPACITY);
                    this.hostPerfMetricsMap.put(host, hostBuffer);
                    // ClusterManagerImpl.logger.info("Added Perf circular buffer for host "
                    // + host.toString());
                }
                hostBuffer.addItem(hostUsage);
                // VirtualMachineManager.log.info("Added host usage " +
                // hostUsage);
            }
        }
    }

    @Override
    public PerfMetricBase[] getVirtualMachinePerfMetrics(final String vmLabel, final String metricId, final Date startTime,
        final Date endTime) throws VMMException {
        if (!metricId.equals("cpu.load") && !metricId.equals("memory.used") && !metricId.equals("network.traffic")
            && !metricId.equals("disk.traffic")) {
            throw new VMMException("InvalidPerfMetricIdException");
        }
        boolean isComposite = false;
        if (metricId.equals("network.traffic") || metricId.equals("disk.traffic")) {
            isComposite = true;
        }
        // VirtualMachineManager.log.info("Start time=" + startTime);
        CircularBuffer<ResourceUsage> buffer = this.virtualMachinePerfMetricsMap.get(vmLabel);
        if (buffer == null) {
            ClusterManagerImpl.logger.info("Cannot find perf data for VM " + vmLabel);
            return new PerfSimpleMetric[0];
        }
        List<PerfMetricBase> metrics = new ArrayList<PerfMetricBase>();
        synchronized (buffer) {
            for (ResourceUsage ru : buffer) {
                Date timeStamp = ru.getSamplingTime();
                // VirtualMachineManager.log.info("Sample time=" + timeStamp);
                if (startTime == null || timeStamp.after(startTime)) {
                    if (endTime == null || timeStamp.before(endTime)) {
                        PerfMetricBase metric = null;
                        if (metricId.equals("cpu.load")) {
                            metric = new PerfSimpleMetric(timeStamp, ru.getCpuLoad());
                        } else if (metricId.equals("memory.used")) {
                            metric = new PerfSimpleMetric(timeStamp, ru.getMemoryUsedKBytes());
                        } else if (metricId.equals("network.traffic")) {
                            if (ru.getNetworkStats().size() > 0) {
                                NetworkStats netStats = ru.getNetworkStats().get(0);
                                metric = new PerfCompositeMetric(timeStamp, new float[] {netStats.getNetReceivedKbitPerSec(),
                                    netStats.getNetTransmittedKbitPerSec()}, new String[] {"in", "out"});
                            }
                        } else if (metricId.equals("disk.traffic")) {
                            if (ru.getDiskStats().size() > 0) {
                                DiskStats diskStats = ru.getDiskStats().get(0);
                                metric = new PerfCompositeMetric(timeStamp, new float[] {diskStats.getDiskReadKBytePerSec(),
                                    diskStats.getDiskWrittenKBytesPerSec()}, new String[] {"in", "out"});
                            }
                        }
                        if (metric != null) {
                            metrics.add(metric);
                        }
                    }
                }
            }
        }
        return metrics.toArray(isComposite ? new PerfCompositeMetric[0] : new PerfSimpleMetric[0]);
    }

    @Override
    public String[] getVirtualMachineRefByHost(final String hostId) {
        this.checkConnection();
        ObjectName objectName;
        try {
            objectName = new ObjectName(hostId);
        } catch (MalformedObjectNameException ex) {
            ClusterManagerImpl.logger.error("Wrong ObjectName: " + hostId, ex);
            return new String[0];
        }
        HostMXBean host = this.getHostMBean(objectName);
        if (host == null) {
            ClusterManagerImpl.logger.error("Cannot retrieve MXBean for host " + hostId);
            return new String[0];
        }
        try {
            List<VirtualMachineMXBean> vms = host.getResidentVMs();
            String result[] = new String[vms.size()];
            for (int i = 0; i < vms.size(); i++) {
                result[i] = vms.get(i).getUuid();
            }
            return result;
        } catch (VMMException ex) {
            ClusterManagerImpl.logger.error("Failed to retrieve VM from host" + hostId, ex);
            return new String[0];
        }
    }

    @Override
    public Map<String, Long> getVirtualMachineSchedulingParams(final String providerId) {
        HashMap<String, Long> result = new HashMap<String, Long>();
        VirtualMachineMXBean vmMBean = this.getVirtualMachineMBean(providerId);
        if (vmMBean != null) {
            try {
                result.put("cap", Long.valueOf(vmMBean.getSchedulingCap()));
                result.put("weight", Long.valueOf(vmMBean.getSchedulingCap()));
            } catch (Exception ex) {
            }
        }
        return result;
    }

    @Override
    public void migrateVirtualMachine(final String providerId, final String destinationHostId) throws VMMException {
        this.checkConnection();
        VirtualMachineMXBean vmMBean = this.getVirtualMachineMBean(providerId);
        if (vmMBean == null) {
            ClusterManagerImpl.logger.info("migrateVM: cannot find VM with given providerId");
            throw new VMMException("VM not found");
        }
        ObjectName objectName;
        try {
            objectName = new ObjectName(destinationHostId);
        } catch (MalformedObjectNameException ex) {
            ClusterManagerImpl.logger.error("Wrong ObjectName: " + destinationHostId, ex);
            throw new VMMException("Invalid host id", ex);
        }
        HostMXBean targetHost = this.getHostMBean(objectName);
        if (targetHost == null) {
            ClusterManagerImpl.logger.error("Cannot retrieve MXBean for host " + destinationHostId);
            throw new VMMException("Host not found");
        }
        vmMBean.migrate(targetHost, true);
    }

    @Override
    public void pauseVirtualMachine(final String providerId) throws VMMException {
        this.checkConnection();
        VirtualMachineMXBean vmMBean = this.getVirtualMachineMBean(providerId);
        if (vmMBean != null) {
            vmMBean.pause();
        } else {
            throw new VMMException("VM not found");
        }
    }

    @Override
    public void rebootVirtualMachine(final String providerId) throws VMMException {
        this.checkConnection();
        VirtualMachineMXBean vmMBean = this.getVirtualMachineMBean(providerId);
        if (vmMBean != null) {
            vmMBean.reboot();
        } else {
            throw new VMMException("VM not found");
        }
    }

    @Override
    public void startVirtualMachine(final String providerId) throws VMMException {
        this.checkConnection();
        VirtualMachineMXBean vmMBean = this.getVirtualMachineMBean(providerId);
        if (vmMBean != null) {
            vmMBean.start();
        } else {
            throw new VMMException("VM not found");
        }
    }

    @Override
    public void stopVirtualMachine(final String providerId) throws VMMException {
        this.checkConnection();
        VirtualMachineMXBean vmMBean = this.getVirtualMachineMBean(providerId);
        if (vmMBean != null) {
            vmMBean.shutdown();
        } else {
            throw new VMMException("VM not found");
        }
    }

    @Override
    public void unpauseVirtualMachine(final String providerId) throws VMMException {
        this.checkConnection();
        VirtualMachineMXBean vmMBean = this.getVirtualMachineMBean(providerId);
        if (vmMBean != null) {
            vmMBean.unpause();
        } else {
            throw new VMMException("VM not found");
        }
    }

    public synchronized void initMonitoringConfig() {
        try {
            // File file = new File(ClusterManagerImpl.PATH_CONFIG_MONITORING);
            // File f = new File(".");
            SAXParser parser = SAXParserFactory.newInstance().newSAXParser();
            ConfigBuilder cb = new ConfigBuilder();
            InputStream is = this.getClass().getResourceAsStream(ClusterManagerImpl.PATH_CONFIG_MONITORING);
            // ClusterManagerImpl.logger.info("Reading resource description file current "
            // + f.getAbsolutePath());
            // ClusterManagerImpl.logger.info("Reading resource description file "
            // + file.getAbsolutePath());
            parser.parse(is, cb);
            this.monitoring_status = cb.isStatus();
            this.hostPerfMetricInfos = cb.getListPerfMetricInfosHost();
            this.virtualMachinePerfMetricInfos = cb.getListPerfMetricInfosVirtualMachine();
            this.serverPoolPerfMetricInfos = cb.getListPerfMetricInfosServerPool();
            this.domainPerfMetricInfos = cb.getListPerfMetricInfosDomain();
            this.userPerfMetricInfos = cb.getListPerfMetricInfosUser();
            this.projectPerfMetricInfos = cb.getListPerfMetricInfosProject();
            this.monitoring_period = cb.getPeriod();
            this.monitoring_delay = cb.getDelay();

        } catch (Exception ex) {
            ClusterManagerImpl.logger.error("While parsing resource description file:", ex);
        }
    }

    @Override
    public PerfMetricInfo[] getVirtualMachinePerfMetricInfos() {
        ClusterManagerImpl.logger.info("called");
        if (this.virtualMachinePerfMetricInfos == null) {
            ClusterManagerImpl.logger.info("init");
            this.initMonitoringConfig();
        }
        return this.virtualMachinePerfMetricInfos;
    }

    @Override
    public PerfMetricInfo[] getHostPerfMetricInfos() {
        ClusterManagerImpl.logger.info("called");
        if (this.hostPerfMetricInfos == null) {
            ClusterManagerImpl.logger.info("init");
            this.initMonitoringConfig();
        }
        return this.hostPerfMetricInfos;
    }

    @Override
    public PerfMetricInfo[] getServerPoolPerfMetricInfos() {
        ClusterManagerImpl.logger.info("called");
        if (this.serverPoolPerfMetricInfos == null) {
            ClusterManagerImpl.logger.info("init");
            this.initMonitoringConfig();
        }
        return this.serverPoolPerfMetricInfos;
    }

    @Override
    public PerfMetricInfo[] getDomainPerfMetricInfos() {
        ClusterManagerImpl.logger.info("called");
        if (this.domainPerfMetricInfos == null) {
            ClusterManagerImpl.logger.info("init");
            this.initMonitoringConfig();
        }
        return this.domainPerfMetricInfos;
    }

    @Override
    public PerfMetricInfo[] getUserPerfMetricInfos(final String id) {
        ClusterManagerImpl.logger.info("called");
        if (this.userPerfMetricInfos == null) {
            ClusterManagerImpl.logger.info("init");
            this.initMonitoringConfig();
        }
        return this.userPerfMetricInfos;
    }

    @Override
    public PerfMetricInfo[] getProjectPerfMetricInfos(final String id) {
        ClusterManagerImpl.logger.info("called");
        if (this.projectPerfMetricInfos == null) {
            ClusterManagerImpl.logger.info("init");
            this.initMonitoringConfig();
        }
        return this.projectPerfMetricInfos;
    }

    @Override
    public PerfMetric getVirtualMachinePerfMetric(final String id, final PerfMetricInfo pmitmp) {
        ClusterManagerImpl.logger.info("called");
        try {
            VirtualMachineMXBean vm = this.getVirtualMachineMBean(id);
            if (vm != null) {
                return vm.getPerfMetric(pmitmp.getName());
            }
            return null;
        } catch (VMMException e) {
            ClusterManagerImpl.logger.error(e);
            return null;
        }
    }

    @Override
    public PerfMetric getHostPerfMetric(final String id, final PerfMetricInfo info) {
        ClusterManagerImpl.logger.info("called");
        this.checkConnection();
        if (id == null) {
            ClusterManagerImpl.logger.error("Host doesn't exist (id=null)");
            return null;
        }
        try {
            HostMXBean host = this.getHostMBean(id);
            if (host != null) {
                return host.getPerfMetric(info.getName());
            }
        } catch (Exception e) {
            ClusterManagerImpl.logger.error("Host doesn't exist :", e);
            return null;
        }
        return null;
    }

    @Override
    public PerfMetric getServerPoolPerfMetric(final String id, final PerfMetricInfo info) {
        ClusterManagerImpl.logger.info("called");
        this.checkConnection();
        if (id == null) {
            ClusterManagerImpl.logger.error("Server pool doesn't exist (id=null)");
            return null;
        }
        try {
            ServerPoolMXBean sp = this.getServerPoolMBean2(id);
            if (sp != null) {
                return sp.getPerfMetric(info.getName());
            }
        } catch (Exception e) {
            ClusterManagerImpl.logger.error("Server pool doesn't exist :", e);
            return null;
        }
        return null;
    }

    @Override
    public PerfMetric getDomainPerfMetric(final String id, final PerfMetricInfo info) {
        ClusterManagerImpl.logger.info("called");
        this.checkConnection();

        PerfMetric result = null;
        int cpt = 0;
        for (ServerPool sp : this.getServerPools()) {
            if (cpt == 0) {
                result = this.getServerPoolPerfMetric(sp.getName(), info);
            } else {
                result.setValue((result.getValue() + this.getServerPoolPerfMetric(id, info).getValue()) / 2);
                // TODO : modify it if you want that metric is a total or an
                // average
            }
        }
        return result;
    }

    @Override
    public PerfMetric[] getHostPerfMetrics(final String id, final PerfMetricInfo info, final Date startTime, final Date endTime) {
        ClusterManagerImpl.logger.info("called");
        this.checkConnection();
        if (id == null) {
            ClusterManagerImpl.logger.error("Host doesn't exist (id=null)");
            return null;
        }
        try {
            HostMXBean host = this.getHostMBean(id);
            if (host != null) {
                ConsolidationFunction cf = ConsolidationFunction.fromValue(info.getType());
                if (cf != null) {
                    return host.getPerfMetrics(info.getName(), startTime, endTime, info.getInterval(), cf);
                }
            }
        } catch (Exception e) {
            ClusterManagerImpl.logger.error("Host doesn't exist :", e);
            return null;
        }
        return null;
    }

    @Override
    public PerfMetric[] getServerPoolPerfMetrics(final String id, final PerfMetricInfo info, final Date startTime,
        final Date endTime) {
        ClusterManagerImpl.logger.info("called");
        this.checkConnection();
        if (id == null) {
            ClusterManagerImpl.logger.error("Server pool doesn't exist (id=null)");
            return null;
        }
        try {
            ServerPoolMXBean sp = this.getServerPoolMBean(id);

            if (sp != null) {
                ConsolidationFunction cf = ConsolidationFunction.fromValue(info.getType());
                if (cf != null) {
                    return sp.getPerfMetrics(info.getName(), startTime, endTime, info.getInterval(), cf);
                }
            }
        } catch (Exception e) {
            ClusterManagerImpl.logger.error("Server pool doesn't exist :", e);
            return null;
        }
        return null;
    }

    @Override
    public PerfMetric[] getDomainPerfMetrics(final String id, final PerfMetricInfo info, final Date startTime,
        final Date endTime) {
        ClusterManagerImpl.logger.info("called");
        this.checkConnection();
        if (id == null) {
            ClusterManagerImpl.logger.error("Domain doesn't exist (id=null)");
            return null;
        }

        return null;
    }

    @Override
    public PerfMetricInfo getPerfMetricInfoById(final Integer valueOf, final String type) {
        System.out.println(type + ": VALUE OF:" + valueOf);
        if (type.equals("USER")) {
            return this.userPerfMetricInfos[valueOf];
        }
        if (type.equals("PROJECT")) {
            return this.projectPerfMetricInfos[valueOf];
        }
        if (type.equals("HOST")) {
            return this.hostPerfMetricInfos[valueOf];
        }
        return null;
    }
}
