package org.ow2.sirocco.cloudmanager.clustermanager.impl;

import java.util.ArrayList;
import java.util.HashMap;

import org.ow2.sirocco.cloudmanager.clustermanager.api.PerfMetricInfo;
import org.ow2.sirocco.vmm.api.monitoring.PerfMetricInfo.Unit;
import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;

public class ConfigBuilder extends DefaultHandler {
    private static Log logger = LogFactory.getLog(ConfigBuilder.class);

    private long period = 300; // by default, period = 5 mins (5*60secs)

    private long delay = 300;

    private boolean status = false;

    private int cpt = 0;

    private ArrayList<String> types = new ArrayList<String>();

    private ArrayList<Long> intervals = new ArrayList<Long>(); // <type
                                                               // (AVERAGE,MIN,MAX...)
                                                               // ,
                                                               // interval>

    private HashMap<String, ArrayList<PerfMetricInfo>> listPerfMetricInfos = new HashMap<String, ArrayList<PerfMetricInfo>>();

    private String entity = null;

    @Override
    public void startElement(final String uri, final String localName, final String qName, final Attributes attrs) {
        if (qName.equals("monitoring")) {
            String status = attrs.getValue("status");
            if (status == null || !status.equals("yes")) {
                ConfigBuilder.logger.error("Missing monitoring name or Monitoring stats <> yes, monitoring no actif");
                return;
            } else {
                this.status = true;
                if (attrs.getValue("period") == null) {
                    ConfigBuilder.logger.error("Missing period, period by default = 300 secs");
                    return;
                } else {
                    this.period = Long.valueOf(attrs.getValue("period"));
                }
                if (attrs.getValue("delay") == null) {
                    ConfigBuilder.logger.error("Missing delay, delay by default = 300 secs");
                    return;
                } else {
                    this.delay = Long.valueOf(attrs.getValue("delay"));
                }
            }
        } else if (qName.equals("storage")) {
            if (attrs.getValue("type") != null && attrs.getValue("interval") != null) {
                this.types.add(attrs.getValue("type"));
                this.intervals.add(Long.valueOf(attrs.getValue("interval")));
            }
        } else if (qName.equals("domain")) {
            this.cpt = 0;
            this.entity = "domain";
            this.listPerfMetricInfos.put(this.entity, new ArrayList<PerfMetricInfo>());
        } else if (qName.equals("serverpool")) {
            this.cpt = 0;
            this.entity = "serverpool";
            this.listPerfMetricInfos.put(this.entity, new ArrayList<PerfMetricInfo>());
        } else if (qName.equals("user")) {
            this.cpt = 0;
            this.entity = "user";
            this.listPerfMetricInfos.put(this.entity, new ArrayList<PerfMetricInfo>());
        } else if (qName.equals("project")) {
            this.cpt = 0;
            this.entity = "project";
            this.listPerfMetricInfos.put(this.entity, new ArrayList<PerfMetricInfo>());
        } else if (qName.equals("host")) {
            this.cpt = 0;
            this.entity = "host";
            this.listPerfMetricInfos.put(this.entity, new ArrayList<PerfMetricInfo>());
        } else if (qName.equals("virtualmachine")) {
            this.cpt = 0;
            this.entity = "virtualmachine";
            this.listPerfMetricInfos.put(this.entity, new ArrayList<PerfMetricInfo>());
        } else if (qName.equals("metric")) {
            if (attrs.getValue("name") != null && attrs.getValue("unit") != null && attrs.getValue("description") != null) {
                for (int i = 0; i < this.types.size(); i++) {
                    PerfMetricInfo info = new PerfMetricInfo();
                    info.setId(this.cpt++);
                    info.setName(attrs.getValue("name"));
                    info.setUnit(Unit.fromValue(attrs.getValue("unit")));
                    info.setDescription(attrs.getValue("description"));
                    info.setType(this.types.get(i));
                    info.setInterval(this.intervals.get(i));
                    this.listPerfMetricInfos.get(this.entity).add(info);
                }
            }
        }
    }

    @Override
    public void endElement(final java.lang.String uri, final java.lang.String localName, final java.lang.String qName)
        throws SAXException {

        if (qName.equals("serverpool") || qName.equals("host") || qName.equals("virtualmachine") || qName.equals("user")
            || qName.equals("project") || qName.equals("domain")) {
            this.entity = null;
        }
    }

    public long getPeriod() {
        return this.period;
    }

    public PerfMetricInfo[] getListPerfMetricInfosHost() {
        if (this.listPerfMetricInfos != null) {
            return this.listPerfMetricInfos.get("host")
                .toArray(new PerfMetricInfo[this.listPerfMetricInfos.get("host").size()]);
        } else {
            return null;
        }
    }

    public PerfMetricInfo[] getListPerfMetricInfosServerPool() {
        if (this.listPerfMetricInfos != null) {

            return this.listPerfMetricInfos.get("serverpool").toArray(
                new PerfMetricInfo[this.listPerfMetricInfos.get("serverpool").size()]);
        } else {
            return null;
        }
    }

    public PerfMetricInfo[] getListPerfMetricInfosDomain() {
        if (this.listPerfMetricInfos != null) {
            return this.listPerfMetricInfos.get("domain").toArray(
                new PerfMetricInfo[this.listPerfMetricInfos.get("domain").size()]);
        } else {
            return null;
        }
    }

    public PerfMetricInfo[] getListPerfMetricInfosVirtualMachine() {
        if (this.listPerfMetricInfos != null) {
            return this.listPerfMetricInfos.get("virtualmachine").toArray(
                new PerfMetricInfo[this.listPerfMetricInfos.get("virtualmachine").size()]);
        } else {
            return null;
        }
    }

    public PerfMetricInfo[] getListPerfMetricInfosUser() {
        if (this.listPerfMetricInfos != null) {
            return this.listPerfMetricInfos.get("user")
                .toArray(new PerfMetricInfo[this.listPerfMetricInfos.get("user").size()]);
        } else {
            return null;
        }
    }

    public PerfMetricInfo[] getListPerfMetricInfosProject() {
        if (this.listPerfMetricInfos != null) {
            return this.listPerfMetricInfos.get("project").toArray(
                new PerfMetricInfo[this.listPerfMetricInfos.get("project").size()]);
        } else {
            return null;
        }
    }

    public boolean isStatus() {
        return this.status;
    }

    public void setDelay(final long delay) {
        this.delay = delay;
    }

    public long getDelay() {
        return this.delay;
    }
}
